from odetam import DetaModel as odeta_model
from datetime import datetime
import ujson
from typing import List, Optional

DETA_BASIC_TYPES = [dict, list, str, int, float, bool]
DETA_OPTIONAL_TYPES = [Optional[type_] for type_ in DETA_BASIC_TYPES]
DETA_BASIC_LIST_TYPES = [
    List[type_] for type_ in DETA_BASIC_TYPES + DETA_OPTIONAL_TYPES
]
DETA_TYPES = DETA_BASIC_TYPES + DETA_OPTIONAL_TYPES + DETA_BASIC_LIST_TYPES

class DetaModel(odeta_model):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    # overwriting odetam's implementation
    def _serialize(self, exclude=None):
        if not exclude:
            exclude = []
        as_dict = {}
        for field_name, field in self.__class__.__fields__.items():
            if field_name == "key" and not self.key:
                continue
            if field_name in exclude:
                continue
            # this originally failed when 0, 0.0, or False. Now we only check for None instances
            if getattr(self, field_name, None) is None:
                as_dict[field_name] = None
                continue
            if field.type_ in DETA_TYPES:
                as_dict[field_name] = getattr(self, field_name)
            elif field.type_ == datetime.datetime:
                as_dict[field_name] = getattr(self, field_name).timestamp()
            elif field.type_ == datetime.date:
                as_dict[field_name] = int(getattr(self, field_name).strftime("%Y%m%d"))
            elif field.type_ == datetime.time:
                as_dict[field_name] = int(
                    getattr(self, field_name).strftime("%H%M%S%f")
                )
            else:
                as_dict[field_name] = ujson.loads(self.json(include={field_name}))[
                    field_name
                ]
        return as_dict
    
    # overwriting odetam's implementation
    @classmethod
    def _deserialize(cls, data):
        as_dict = {}
        for field_name, field in cls.__fields__.items():
            # this originally failed when 0, 0.0, or False. Now we only check for None instances
            if field_name not in data:
                continue
            if field.type_ in DETA_TYPES:
                as_dict[field_name] = data[field_name]
            elif field.type_ == datetime.datetime:
                as_dict[field_name] = datetime.datetime.fromtimestamp(data[field_name])
            elif field.type_ == datetime.date:
                as_dict[field_name] = datetime.datetime.strptime(
                    str(data[field_name]), "%Y%m%d"
                ).date()
            elif field.type_ == datetime.time:
                as_dict[field_name] = datetime.datetime.strptime(
                    str(data[field_name]), "%H%M%S%f"
                ).time()
            else:
                try:
                    as_dict[field_name] = ujson.loads(data.get(field_name))
                except (TypeError, ValueError):
                    as_dict[field_name] = data.get(field_name, None)
        return cls.parse_obj(as_dict)


from fastapi import APIRouter, Request
from detamvc.model import DetaModel

from fastapi.templating import Jinja2Templates


{obj}_router = APIRouter()
templates = Jinja2Templates(directory="")


class {Obj}(DetaModel):
    {model_attrs}

    class Config:
        table_name = "{proj}_{obj}"

# INDEX
@{obj}_router.get('/')
def index(request: Request):
    {obj}_list = {Obj}.get_all()
    return templates.TemplateResponse(
        '{obj}/templates/index.html',
        context={{'request': request, '{obj}_list': {obj}_list }})

# CREATE
@{obj}_router.get('/new')
def new(request: Request):
    return templates.TemplateResponse(
        '{obj}/templates/form.html',
        context={{'request': request, 'vals': dict() }})

@{obj}_router.post('/new', response_model={Obj})
async def create(request: Request):
    form_data = await request.form()
    {obj} = {Obj}.parse_obj(form_data)
    {obj}.save()
    return view(request, {obj}.key)

# UPDATE
@{obj}_router.get('/edit/{{key}}')
def edit(request: Request, key: str):
    vals = {Obj}.get(key)
    return templates.TemplateResponse(
        '{obj}/templates/form.html',
        context={{'request': request, 'vals': vals.dict() }})

@{obj}_router.post("/edit/{{key}}")
async def update(request: Request, key: str):
    stored_data = {Obj}.get(key)
    update_data = await request.form()
    parsed = {Obj}.parse_obj(update_data).dict(exclude_unset=True)
    parsed['key'] = key
    updated = stored_data.copy(update=parsed)
    updated.save()
    return view(request, key)

# DELETE
@{obj}_router.get('/delete/{{key}}')
def delete(request: Request, key: str):
    {Obj}.delete_key(key)
    return index(request)

# VIEW
@{obj}_router.get('/{{key}}')
def view(request: Request, key: str):
    {obj} = {Obj}.get(key)
    return templates.TemplateResponse(
        '{obj}/templates/view.html',
        context={{'request': request, '{obj}': {obj} }})




import json 
from pathlib import Path

CONFIG_PATH = Path(__file__).parent.resolve() / 'user_config.json'

def __fetch_config():
    with open(CONFIG_PATH, 'r') as j:
        return json.loads(j.read())

def set_project_key(project_key: str):
    conf = __fetch_config()
    conf['project_key'] = project_key 
    with open(CONFIG_PATH, 'w') as j:
        j.write(json.dumps(conf, indent=4))

def clear_project_key():
    set_project_key("")

def config(key: str or None = None):
    if key:
        return __fetch_config().get(key)
    return __fetch_config()



import typer
import os
from detamvc import __version__
from .generator import build_base, gen_scaffold
from typing import List
from . import utilities as utils

app = typer.Typer()


@app.command()
def version():
    """ show current version of DetaMVC"""
    typer.echo(f"DetaMVC {__version__}")

@app.command()
def new(project: str):
    """ create a new project """
    new_folder = os.path.join(os.curdir, project)
    try:      
        typer.secho(f"\nBuilding new project: {project}\n", fg='green')
        build_base(new_folder, project)
        typer.echo(f"\n{project} was created.\n")
    except FileExistsError:
        typer.secho(f"'{project}' already exists in this folder.\n", fg='red')

@app.command()
def server():
    """ run the app locally """
    os.system("uvicorn main:app --reload")

@app.command()
def s():
    """ alias for 'server' """
    server()

@app.command()
def scaffold(obj: str, attributes: List[str]):
    """ create a router and views for a described object """
    typer.secho(f"\nScaffolding views and router for: {obj}\n", fg='green')
    gen_scaffold(os.curdir, obj, attributes)
    typer.echo(f"\n{obj} was created.\n")

@app.command()
def set_project_key(project_key: str):
    """ set your development project key for Deta for faster builds """
    utils.set_project_key(project_key)
    typer.secho("Project Key successfully set.", fg='green')

@app.command()
def clear_project_key():
    """ clear your Deta development project key """
    utils.clear_project_key()
    typer.secho("Project Key successfully cleared.", fg='green')

@app.command()
def view_config():
    """ view your development configurations """
    conf = utils.config()
    typer.echo(conf)




from pathlib import Path
from os import walk, makedirs
from .utilities import config


def __builder(project_path, generator_path, format_attrs: dict, scaffold_obj_name=''):
    project_path = Path(project_path)
    generator_path = Path(__file__).parent.resolve() / generator_path
    for (dirpath, _, filenames) in walk(generator_path):
        np = (Path(scaffold_obj_name) / project_path 
              if scaffold_obj_name else project_path)
        new_path = dirpath.replace(str(generator_path), str(np))
        if '__pycache__' in str(dirpath):
            continue
        makedirs(new_path, exist_ok=True)
        for filename in filenames:
            full_path = Path(dirpath) / filename
            print(f">> {Path(new_path) / filename}")
            with open(full_path, 'r') as c:
                content = c.read()          
            
            with open(Path(new_path) / filename, 'w+') as o:
                o.write(__format(content, format_attrs))


def build_base(p, project_name):
    makedirs(p)
    content_attrs = {
        'Obj': project_name.title(),
        'project_key': config('project_key') or ''}
    __builder(
        project_path=p, 
        generator_path='templates/core', 
        format_attrs=content_attrs)


def gen_scaffold(p, obj, attributes):
    obj_attrs = __get_object_attrs(attributes)
    model_attrs = '\n    '.join([f"{k}: {'str' if v == 'text' else v}" for k, v in obj_attrs.items()])
    form_attrs = __create_form_attrs(
        attributes=obj_attrs, 
        helpers_path='templates/scaffold_helpers')
    format_attrs = {
        'proj': Path.cwd().name,
        'obj': obj, 
        'Obj': obj.title(), 
        'model_attrs': model_attrs, 
        'form_attrs': form_attrs}
    __builder(
        project_path=p, 
        generator_path='templates/scaffold', 
        format_attrs=format_attrs,
        scaffold_obj_name=obj)

    update_main(p, obj)


def update_main(p, name):
    main_file = Path(p) / 'main.py'
    with open(main_file, 'r') as m:
            main = m.readlines()
    imp = f'from {name}.router import {name}_router\n'
    inc = f'app.include_router({name}_router, tags=["{name}"], prefix="/{name}")\n'
    imp_placement = 0
    for i, line in enumerate(main):
        if '_router' in line and 'import' in line:
            imp_placement = i + 1
    main.insert(imp_placement, imp)
    main.append(inc)

    with open(main_file, 'w') as m:
        m.write(''.join(main))


def __format(doc, attr_dict):
    built_attrs = {}
    for k, v in attr_dict.items():
        if f'{{{k}}}' in doc:
            built_attrs.update({k: v})
    if built_attrs:
        doc = doc.format(**built_attrs)
    return doc


def __get_object_attrs(attributes):
    obj_attrs = dict()
    for a in attributes:
        if not a:
            continue
        s = a.split(':')
        obj_attrs[s[0]] = 'str' if len(s) == 1 else s[1].lower()
    return obj_attrs
        

def __create_form_attrs_orig(attributes):
    # TODO: enhance the forms created for various data types
    # textarea, color, date...
    fa = list()
    for f, t in attributes.items():

        fa.append(f"""<label for="{f}">{f.title()}</label><br>""")

        # process datatypes as provided. Defaults to 'str' -> 'text'
        ty = "text"
        if t in ('int', 'float'):
            ty = "number"
        elif t == 'text':
            ty = 'textarea'
        elif t == 'bool':
            ty = 'bool-select'

        # create corresponding field for input based on type
        if ty in ("text", "number"):
            fld = f"""<input type="{ty}" name="{f}" class="{'form-check-input' if ty == 'checkbox' else 'form-control'}" value="{{{{vals.get('{f}', '')}}}}"{ ' step="0.01"' if t == 'float' else ''}><br>"""
        elif ty == 'textarea':
            fld = f"""<textarea name="{f}" class="form-control">{{{{vals.get('{f}', '')}}}}</textarea><br>"""
        elif ty == 'bool-select':
            fld = f"""<select name="{f}" class="form-select">
            <option value="true">True</option>
            <option value="false">False</option>
            </select>"""

        fa.append(fld)
    return '\n\t\t\t\t\t\t'.join(fa)

def __create_form_attrs(attributes, helpers_path):
    helpers_path = Path(__file__).parent.resolve() / helpers_path
    form_fields = list()
    for f, t in attributes.items():
        field_helper = helpers_path / f'{t}.html'
        if not field_helper.exists():
            field_helper = helpers_path / 'str.html'
        with open(field_helper, 'r') as o:
            content = o.read()
        format_attrs = {'f': f, 'F': f.title()}
        form_fields.append(__format(content, format_attrs))
    return '\n'.join(form_fields)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['detamvc',
 'detamvc.templates.core',
 'detamvc.templates.core.static_pages',
 'detamvc.templates.scaffold']

package_data = \
{'': ['*'],
 'detamvc': ['templates/core/static_pages/static/css/*',
             'templates/core/static_pages/static/img/*',
             'templates/core/static_pages/templates/*',
             'templates/scaffold/templates/*',
             'templates/scaffold_helpers/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'deta>=1.1.0,<2.0.0',
 'fastapi>=0.78.0,<0.79.0',
 'odetam>=1.2.0,<2.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'python-multipart>=0.0.5,<0.0.6',
 'typer[all]>=0.4.1,<0.5.0']

entry_points = \
{'console_scripts': ['detamvc = detamvc.main:app']}

setup_kwargs = {
    'name': 'detamvc',
    'version': '0.1.8',
    'description': '',
    'long_description': '# DetaMVC\n\n\nDetaMVC is a framework for rapidly developing web applications using:\n- [FastAPI](https://fastapi.tiangolo.com/)\n- [Jinja2](https://fastapi.tiangolo.com/advanced/templates/?h=jinja2)\n- [Deta](https://docs.deta.sh/docs/home)\n- [ODetaM](https://github.com/rickh94/ODetaM)\n\n## Work in Progress\nPlease note that DetaMVC is a work in progress currently. Some planned updates include:\n- Support for more data types when scaffolding\n- Implement Deta Auth when available\n- Generate with other front ends ? (may take significant restructuring)\n\n## Installation\n```\npip install detamvc\n```\n   \n\n## Basics\nIf you are familiar with Ruby on Rails, the commands are very similar for creating an application. \n\n```\ndetamvc new project\n\ncd project\n\ndetamvc scaffold post title:str content:str\n```\n\nBefore running your project, be sure to set your PROJECT_KEY for Deta. You can get this from your dashboard under \'settings\'.\n\n\n```\necho PROJECT_KEY="#######_#############" > .env\n```\n\n**NOTE**: Deta Base is used as your database. I would strongly recommend creating a new Project called \'development\' where you can play around with ideas while building. When you go to production, create a new project in Deta for this to be managed in.\n\n## Run a Server Manually\n\nThis assumes you have uvicorn installed. You can run with other servers as you wish - just set up like you would for a normal [FastAPI](https://fastapi.tiangolo.com/deployment/manually/ "Run a Server Manually - Uvicorn") application.\n```\ndetamvc s\n```\nor\n```\nuvicorn main:app --reload\n```\n\n## Deploy on Deta\nNow you can deploy this on Deta!\n```\ndeta new --project default\n```',
    'author': 'MBeebe',
    'author_email': 'grow.food.everywhere@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
